import os, strutils
import nigui, nigui/msgbox
import ripen, drawing

const AppTitle = "Tipsy Turtle"

const AboutMessage = """
Tipsy Turtle 1.0.2 beta
procedural graphics tool
by No Time To Play

21 March 2023; MIT License
"""

app.init

var top = newWindow(AppTitle)

top.width = 854.scaleToDpi
top.height = 480.scaleToDpi

var layout = newLayoutContainer(Layout_Vertical)

layout.widthMode = WidthMode_Fill
layout.heightMode = HeightMode_Fill

var toolbar = newLayoutContainer(Layout_Horizontal)

toolbar.widthMode = WidthMode_Fill

var toNew = newButton("New")
var toOpen = newButton("Open")
var toSave = newButton("Save")
var toResize = newButton("Resize")
var toConsole = newButton("Console")
var toAbout = newButton("About")

var status = newLabel("Welcome to Tipsy Turtle version 1.0.2 beta")

status.widthMode = WidthMode_Fill
status.heightMode = HeightMode_Fill
status.xTextAlign = XTextAlign_Center
status.yTextAlign = YTextAlign_Center

toolbar.add toNew
toolbar.add toOpen
toolbar.add toSave
toolbar.add toResize
toolbar.add toConsole

toolbar.add status

toolbar.add toAbout

var viewport = newControl()

viewport.widthMode = WidthMode_Fill
viewport.heightMode = HeightMode_Fill

var command = newTextBox()

command.widthMode = WidthMode_Fill

layout.add toolbar
layout.add viewport
layout.add command

top.add layout


var askOpen = newOpenFileDialog()
var askSave = newSaveFileDialog()
askSave.title = "Save image"


var console = newWindow("Forth console")

console.width = 640.scaleToDpi
console.height = 400.scaleToDpi

var layout2 = newLayoutContainer(Layout_Vertical)

layout2.widthMode = WidthMode_Fill
layout2.heightMode = HeightMode_Fill

var toolbar2 = newLayoutContainer(Layout_Horizontal)

toolbar2.widthMode = WidthMode_Fill

var toLoad = newButton("Load...")
var toRun = newButton("Run")
var toClear = newButton("Clear")

toolbar2.add toLoad
toolbar2.add toRun
toolbar2.add toClear

var status2 = newLabel("F5 to run")

status2.widthMode = WidthMode_Fill
status2.heightMode = HeightMode_Fill
status2.xTextAlign = XTextAlign_Center
status2.yTextAlign = YTextAlign_Center

toolbar2.add status2

var panes = newLayoutContainer(Layout_Horizontal)

panes.widthMode = WidthMode_Fill
panes.heightMode = HeightMode_Fill

var editor = newTextArea()

editor.widthMode = WidthMode_Fill
editor.heightMode = HeightMode_Fill

var output = newTextArea()

output.widthMode = WidthMode_Fill
output.heightMode = HeightMode_Fill
output.editable = false

panes.add editor
panes.add output

layout2.add toolbar2
layout2.add panes
console.add layout2


var sizer = newWindow("Resize image")

sizer.width = 240.scaleToDpi
sizer.height = 80.scaleToDpi

var layout3 = newLayoutContainer(Layout_Vertical)

layout3.widthMode = WidthMode_Fill
layout3.heightMode = HeightMode_Fill

var fields = newLayoutContainer(Layout_Horizontal)

fields.widthMode = WidthMode_Fill
fields.yAlign = YAlign_Center

var width = newTextBox()
var height = newTextBox()

fields.add newLabel("New size: ")
fields.add width
fields.add newLabel(" x ")
fields.add height

var buttons = newLayoutContainer(Layout_Horizontal)

buttons.widthMode = WidthMode_Fill
buttons.xAlign = XAlign_Right

var toReallyResize = newButton("Resize")
var toCancelResize = newButton("Cancel")

buttons.add toReallyResize
buttons.add toCancelResize

layout3.add fields
layout3.add buttons
sizer.add layout3

proc pstack() = output.addLine stack.join(" ")
proc pwords() =
        for i in words.keys:
                output.addText i & " "
proc palias() =
        for i in words.keys:
                output.addText i & " "
proc psigils() =
        for i in sigil.keys:
                output.addText i & " "
proc pbuffer() = output.addLine buffer.join(" ")

words[".s"] = primitive(pstack)
words["words"] = primitive(pwords)
words["aliases"] = primitive(palias)
words["sigils"] = primitive(psigils)
words["buffer"] = primitive(pbuffer)

sigil['?'] = proc (n: string) = output.addLine $words[n]

proc period() = output.addText formatFloat(stack.pop, precision = -1) & " "
proc pspace() = output.addText " "
proc ptype() = output.addText scratch & " "
proc cr() = output.addLine()

words["."] = primitive(period)
words["space"] = primitive(pspace)
words[".."] = primitive(ptype)
words["cr"] = primitive(cr)

proc alert() = top.alert scratch, "The turtle says:"
proc confirm() =
        stack.add top.msgBox(scratch,
                "The turtle asks:", "Yes", "No").toFloat
proc bye() = app.quit

words["alert"] = primitive(alert)
words["confirm"] = primitive(confirm)
words["bye"] = primitive(bye)

output.addLine "Welcome to " & version_string
output.addLine $words.len & " words and " & $sigil.len & " sigils defined."

var history: seq[string]
var histSize = 100
var histPos = 0

var fileName = ""

proc doNew() =
        let answer = top.msgBox(
                "Drawing isn't saved. Start another?",
                AppTitle, "Start anew", "Keep")
        if answer != 1: return
        
        image.resize image.width, image.height
        image.canvas.areaColor = rgb(0, 0, 0)
        image.canvas.lineColor = rgb(255, 255, 255)
        image.canvas.textColor = rgb(255, 255, 255)

        turtleX = float(image.width div 2)
        turtleY = float(image.height div 2)
        heading = 0.0
        showTurtle = true
        isPenDown = true

        modified = false

        viewport.forceRedraw
        top.title = AppTitle
        fileName = ""

proc doOpen() =
        let answer = top.msgBox(
                "Drawing isn't saved. Open another?",
                AppTitle, "Open", "Keep")
        if answer != 1: return
        
        askOpen.title = "Open image"
        run(askOpen)
        if len(askOpen.files) == 0:
                status.text = "Open canceled."
        else:
                try:
                        image.loadFromFile(askOpen.files[0])
                        let (_, fn, ext) = askOpen.files[0].splitFile
                        top.title = fn & ext & " | " & AppTitle
                        fileName = askOpen.files[0]
                        viewport.forceRedraw
                        modified = false
                except IOError:
                        top.alert(getCurrentExceptionMsg(), AppTitle)
                except Exception:
                        # loadFromFile raises this on unknown formats
                        top.alert(getCurrentExceptionMsg(), AppTitle)
                        image = newImage()
                        setupImage()
                        modified = false

proc doSave() =
        let (dir, fn, ext) = fileName.splitFile
        askSave.directory = dir
        askSave.defaultName = fn & ext
        run(askSave)
        if askSave.file != "":
                try:
                        let (_, fn, ext) = askSave.file.splitFile
                        case ext.toLowerAscii
                        of ".bmp":
                                image.saveToBitmapFile(askSave.file)
                                status.text = "Saved as " & fn & ext
                                top.title = fn & ext & " | " & AppTitle
                                fileName = askOpen.files[0]
                                modified = false
                        of ".png":
                                image.saveToPngFile(askSave.file)
                                status.text = "Saved as " & fn & ext
                                top.title = fn & ext & " | " & AppTitle
                                fileName = askOpen.files[0]
                                modified = false
                        of ".jpg", ".jpeg":
                                image.saveToJpegFile(askSave.file)
                                status.text = "Saved as " & fn & ext
                                top.title = fn & ext & " | " & AppTitle
                                fileName = askOpen.files[0]
                                modified = false
                        else:
                                let msg = "Please save as BMP, PNG or JPG."
                                top.alert(msg, AppTitle)
                except IOError:
                        top.alert(getCurrentExceptionMsg(), AppTitle)
        else:
                status.text = "Save canceled."

toNew.onClick = proc(event: ClickEvent) = doNew()
toOpen.onClick = proc(event: ClickEvent) = doOpen()
toSave.onClick = proc(event: ClickEvent) = doSave()

toConsole.onClick = proc(event: ClickEvent) =
        console.show
        editor.focus
toResize.onClick = proc(event: ClickEvent) =
        width.text = $image.width
        height.text = $image.height
        sizer.show
toAbout.onClick = proc(event: ClickEvent) =
        top.alert(AboutMessage, "About " & AppTitle)

proc doRun() =
        try:
                atLine = 1
                for i in editor.text.splitLines:
                        interpret(i.splitWhitespace)
                        inc atLine
                viewport.forceRedraw
                output.addLine "ok"
        except IndexDefect:
                console.alert "Stack underflow at line " & $atLine, AppTitle
        except CatchableError:
                let msg = getCurrentExceptionMsg()
                console.alert msg & " at line " & $atLine, AppTitle
toLoad.onClick = proc(event: ClickEvent) =
        askOpen.title = "Load script"
        run(askOpen)
        if len(askOpen.files) == 0:
                status2.text = "Load canceled."
        else:
                try:
                        editor.text = readFile(askOpen.files[0])
                except IOError:
                        top.alert(getCurrentExceptionMsg(), AppTitle)
toRun.onClick = proc(event: ClickEvent) = doRun()
toClear.onClick = proc(event: ClickEvent) = output.text = ""

console.onKeyDown = proc(event: KeyboardEvent) =
        case event.key
        of Key_F5:
                doRun()
        of Key_L:
                if Key_ControlL.isDown: editor.focus
        of Key_R:
                if Key_ControlL.isDown: doRun()
        else:
                discard

toReallyResize.onClick = proc(event: ClickEvent) =
        var tmp = newImage()

        tmp.resize parseInt(width.text), parseInt(height.text)
        
        discard tmp.canvas.centerImage image

        tmp.canvas.lineColor = image.canvas.lineColor
        tmp.canvas.textColor = image.canvas.textColor
        
        image = tmp
        viewport.forceRedraw
        sizer.hide
toCancelResize.onClick = proc(event: ClickEvent) = sizer.hide

viewport.onDraw = proc (event: DrawEvent) =
        let canvas = event.control.canvas
        
        let (xOff, yOff) = canvas.centerImage(image)
        let tx = int(turtleX) + xOff
        let ty = int(turtleY) + yOff
        
        canvas.lineWidth = 2.0
        canvas.lineColor = image.canvas.lineColor
        canvas.drawRectOutline(xOff, yOff, image.width, image.height)
        if showTurtle:
                canvas.drawEllipseOutline(tx - 12, ty - 12, 24, 24)
                canvas.drawLineRelative(tx, ty, heading, 12)

command.onKeyDown = proc(event: KeyboardEvent) =
        case event.key
        of Key_Return:
                try:
                        interpret(command.text.splitWhitespace)
                        viewport.forceRedraw
                except IndexDefect:
                        top.alert "Stack underflow", AppTitle
                except CatchableError:
                        top.alert getCurrentExceptionMsg(), AppTitle
                history.add command.text
                if len(history) > histsize:
                        history.delete 0
                histPos = len(history)
                command.text = ""
        of Key_PageUp:
                if histPos > 0:
                        dec histPos
                        command.text = history[histPos]
        of Key_PageDown:
                if histPos < len(history):
                        inc histPos
                        if histPos < len(history):
                                command.text = history[histPos]
                        else:
                                command.text = ""
        else:
                discard

console.onCloseClick = proc(event: CloseClickEvent) = console.hide
sizer.onCloseClick = proc(event: CloseClickEvent) = sizer.hide
top.onCloseClick = proc(event: CloseClickEvent) =
        if modified:
                let answer = top.msgBox(
                        "Drawing isn't saved. Quit anyway?",
                        AppTitle, "Quit", "Stay")
                if answer != 1: return
        app.quit

top.onKeyDown = proc(event: KeyboardEvent) =
        case event.key
        of Key_F1:
                top.alert(AboutMessage, "About " & AppTitle)
        of Key_L:
                if Key_ControlL.isDown: command.focus
        of Key_N:
                if Key_ControlL.isDown: doNew()
        of Key_O:
                if Key_ControlL.isDown: doOpen()
        of Key_Q:
                if Key_ControlL.isDown: app.quit
        of Key_S:
                if Key_ControlL.isDown: doSave()
        else:
                discard

top.show
command.focus
app.run
